/*
 *
 * Name: main.c
 * Author: CClassicVideos@aol.com
 * Creation Date: 19.05.2012
 * Copyright: © 2012 by C_Classic
 * License: GPL V.3 
 *
 * Note: This application is based around the V-USB MIDI-Example by Martin Homuth-Rosemann
 *
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/delay.h>

#include "hardware.h"
#include "usbdrv.h"
#include "uart.h"

#if DEBUG_LEVEL > 0
#	warning "Never compile production devices with debugging enabled"
#endif


uchar key[2], lastKey[2];

uchar midiMsg[8];
uchar iii;

uchar oldADC[8];
uchar ADCidx;
uchar SYSidx;

uint8_t serCode[5];
uint8_t serIdx=0;
uchar newSerial = 0;

// This descriptor is based on http://www.usb.org/developers/devclass_docs/midi10.pdf
// 
// Appendix B. Example: Simple MIDI Adapter (Informative)
// B.1 Device Descriptor
//
static PROGMEM char deviceDescrMIDI[] = {	/* USB device descriptor */
	18,			/* sizeof(usbDescriptorDevice): length of descriptor in bytes */
	USBDESCR_DEVICE,	/* descriptor type */
	0x10, 0x01,		/* USB version supported */
	0,			/* device class: defined at interface level */
	0,			/* subclass */
	0,			/* protocol */
	8,			/* max packet size */
	USB_CFG_VENDOR_ID,	/* 2 bytes */
	USB_CFG_DEVICE_ID,	/* 2 bytes */
	USB_CFG_DEVICE_VERSION,	/* 2 bytes */
	1,			/* manufacturer string index */
	2,			/* product string index */
	0,			/* serial number string index */
	1,			/* number of configurations */
};

// B.2 Configuration Descriptor
static PROGMEM char configDescrMIDI[] = {	/* USB configuration descriptor */
	9,			/* sizeof(usbDescrConfig): length of descriptor in bytes */
	USBDESCR_CONFIG,	/* descriptor type */
	101, 0,			/* total length of data returned (including inlined descriptors) */
	2,			/* number of interfaces in this configuration */
	1,			/* index of this configuration */
	0,			/* configuration name string index */
#if USB_CFG_IS_SELF_POWERED
	USBATTR_SELFPOWER,	/* attributes */
#else
	USBATTR_BUSPOWER,	/* attributes */
#endif
	USB_CFG_MAX_BUS_POWER / 2,	/* max USB current in 2mA units */

// B.3 AudioControl Interface Descriptors
// The AudioControl interface describes the device structure (audio function topology) 
// and is used to manipulate the Audio Controls. This device has no audio function 
// incorporated. However, the AudioControl interface is mandatory and therefore both 
// the standard AC interface descriptor and the classspecific AC interface descriptor 
// must be present. The class-specific AC interface descriptor only contains the header 
// descriptor.

// B.3.1 Standard AC Interface Descriptor
// The AudioControl interface has no dedicated endpoints associated with it. It uses the 
// default pipe (endpoint 0) for all communication purposes. Class-specific AudioControl 
// Requests are sent using the default pipe. There is no Status Interrupt endpoint provided.
	/* AC interface descriptor follows inline: */
	9,			/* sizeof(usbDescrInterface): length of descriptor in bytes */
	USBDESCR_INTERFACE,	/* descriptor type */
	0,			/* index of this interface */
	0,			/* alternate setting for this interface */
	0,			/* endpoints excl 0: number of endpoint descriptors to follow */
	1,			/* */
	1,			/* */
	0,			/* */
	0,			/* string index for interface */

// B.3.2 Class-specific AC Interface Descriptor
// The Class-specific AC interface descriptor is always headed by a Header descriptor 
// that contains general information about the AudioControl interface. It contains all 
// the pointers needed to describe the Audio Interface Collection, associated with the 
// described audio function. Only the Header descriptor is present in this device 
// because it does not contain any audio functionality as such.
	/* AC Class-Specific descriptor */
	9,			/* sizeof(usbDescrCDC_HeaderFn): length of descriptor in bytes */
	36,			/* descriptor type */
	1,			/* header functional descriptor */
	0x0, 0x01,		/* bcdADC */
	9, 0,			/* wTotalLength */
	1,			/* */
	1,			/* */

// B.4 MIDIStreaming Interface Descriptors

// B.4.1 Standard MS Interface Descriptor
	/* interface descriptor follows inline: */
	9,			/* length of descriptor in bytes */
	USBDESCR_INTERFACE,	/* descriptor type */
	1,			/* index of this interface */
	0,			/* alternate setting for this interface */
	2,			/* endpoints excl 0: number of endpoint descriptors to follow */
	1,			/* AUDIO */
	3,			/* MS */
	0,			/* unused */
	0,			/* string index for interface */

// B.4.2 Class-specific MS Interface Descriptor
	/* MS Class-Specific descriptor */
	7,			/* length of descriptor in bytes */
	36,			/* descriptor type */
	1,			/* header functional descriptor */
	0x0, 0x01,		/* bcdADC */
	65, 0,			/* wTotalLength */

// B.4.3 MIDI IN Jack Descriptor
	6,			/* bLength */
	36,			/* descriptor type */
	2,			/* MIDI_IN_JACK desc subtype */
	1,			/* EMBEDDED bJackType */
	1,			/* bJackID */
	0,			/* iJack */

	6,			/* bLength */
	36,			/* descriptor type */
	2,			/* MIDI_IN_JACK desc subtype */
	2,			/* EXTERNAL bJackType */
	2,			/* bJackID */
	0,			/* iJack */

//B.4.4 MIDI OUT Jack Descriptor
	9,			/* length of descriptor in bytes */
	36,			/* descriptor type */
	3,			/* MIDI_OUT_JACK descriptor */
	1,			/* EMBEDDED bJackType */
	3,			/* bJackID */
	1,			/* No of input pins */
	2,			/* BaSourceID */
	1,			/* BaSourcePin */
	0,			/* iJack */

	9,			/* bLength of descriptor in bytes */
	36,			/* bDescriptorType */
	3,			/* MIDI_OUT_JACK bDescriptorSubtype */
	2,			/* EXTERNAL bJackType */
	4,			/* bJackID */
	1,			/* bNrInputPins */
	1,			/* baSourceID (0) */
	1,			/* baSourcePin (0) */
	0,			/* iJack */


// B.5 Bulk OUT Endpoint Descriptors

//B.5.1 Standard Bulk OUT Endpoint Descriptor
	9,			/* bLenght */
	USBDESCR_ENDPOINT,	/* bDescriptorType = endpoint */
	0x1,			/* bEndpointAddress OUT endpoint number 1 */
	3,			/* bmAttributes: 2:Bulk, 3:Interrupt endpoint */
	8, 0,			/* wMaxPacketSize */
	10,			/* bIntervall in ms */
	0,			/* bRefresh */
	0,			/* bSyncAddress */

// B.5.2 Class-specific MS Bulk OUT Endpoint Descriptor
	5,			/* bLength of descriptor in bytes */
	37,			/* bDescriptorType */
	1,			/* bDescriptorSubtype */
	1,			/* bNumEmbMIDIJack  */
	1,			/* baAssocJackID (0) */


//B.6 Bulk IN Endpoint Descriptors

//B.6.1 Standard Bulk IN Endpoint Descriptor
	9,			/* bLenght */
	USBDESCR_ENDPOINT,	/* bDescriptorType = endpoint */
	0x81,			/* bEndpointAddress IN endpoint number 1 */
	3,			/* bmAttributes: 2: Bulk, 3: Interrupt endpoint */
	8, 0,			/* wMaxPacketSize */
	10,			/* bIntervall in ms */
	0,			/* bRefresh */
	0,			/* bSyncAddress */

// B.6.2 Class-specific MS Bulk IN Endpoint Descriptor
	5,			/* bLength of descriptor in bytes */
	37,			/* bDescriptorType */
	1,			/* bDescriptorSubtype */
	1,			/* bNumEmbMIDIJack (0) */
	3,			/* baAssocJackID (0) */
};


uchar usbFunctionDescriptor(usbRequest_t * rq)
{

	if (rq->wValue.bytes[1] == USBDESCR_DEVICE) {
		usbMsgPtr = (uchar *) deviceDescrMIDI;
		return sizeof(deviceDescrMIDI);
	} else {		/* must be config descriptor */
		usbMsgPtr = (uchar *) configDescrMIDI;
		return sizeof(configDescrMIDI);
	}
}

static uchar sendEmptyFrame;


/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

uchar usbFunctionSetup(uchar data[8])
{
	usbRequest_t *rq = (void *) data;

	// DEBUG LED
	//PORTC ^= 0x01;

	if ((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS) {	/* class request type */

		/*  Prepare bulk-in endpoint to respond to early termination   */
		if ((rq->bmRequestType & USBRQ_DIR_MASK) ==
		    USBRQ_DIR_HOST_TO_DEVICE)
			sendEmptyFrame = 1;
	}

	return 0xff;
}


/*---------------------------------------------------------------------------*/
/* usbFunctionRead                                                           */
/*---------------------------------------------------------------------------*/

uchar usbFunctionRead(uchar * data, uchar len)
{
	// DEBUG LED
	//PORTC ^= 0x02;

	data[0] = 0;
	data[1] = 0;
	data[2] = 0;
	data[3] = 0;
	data[4] = 0;
	data[5] = 0;
	data[6] = 0;

	return 7;
}


/*---------------------------------------------------------------------------*/
/* usbFunctionWrite                                                          */
/*---------------------------------------------------------------------------*/

uchar usbFunctionWrite(uchar * data, uchar len)
{
	// DEBUG LED
	//PORTC ^= 0x04;
	return 1;
}


/*---------------------------------------------------------------------------*/
/* usbFunctionWriteOut                                                       */
/*                                                                           */
/* this Function is called if a MIDI Out message (from PC) arrives.          */
/*                                                                           */
/*---------------------------------------------------------------------------*/

void usbFunctionWriteOut(uchar * data, uchar len)
{
	// DEBUG LED
	//PORTC ^= 0x20;
}

void hardware_init() {
	// Reset to default
		DDRA = 0;
		DDRB = 0;
		DDRC = 0;
		DDRD = 0;

	// Setup USB
		uchar i, j;
		/* activate pull-ups except on USB lines */
		USB_CFG_IOPORT =
		    (uchar) ~ ((1 << USB_CFG_DMINUS_BIT) |
			       (1 << USB_CFG_DPLUS_BIT));
		/* all pins input except USB (-> USB reset) */
		#ifdef USB_CFG_PULLUP_IOPORT	/* use usbDeviceConnect()/usbDeviceDisconnect() if available */
		USBDDR = 0;		/* we do RESET by deactivating pullup */
		usbDeviceDisconnect();
		#else
		USBDDR = (1 << USB_CFG_DMINUS_BIT) | (1 << USB_CFG_DPLUS_BIT);
		#endif

		j = 0;
		while (--j) {		/* USB Reset by device only required on Watchdog Reset */
			i = 0;
			while (--i);	/* delay >10ms for USB reset */
		}
		#ifdef USB_CFG_PULLUP_IOPORT
		usbDeviceConnect();
		#else
		USBDDR = 0;		/*  remove USB reset condition */
		#endif

	// LEDs -> Output
		LED_DDR |= (1 << LED_LEFT);
		LED_DDR |= (1 << LED_MIDDLE);
		LED_DDR |= (1 << LED_RIGHT);

	// Buttons
		BTN1_DDR = 0x0; // All inputs
		BTN2_DDR = 0x0; // All inputs
		BTN1_PORT = 0xff; // Pullup
		BTN2_PORT = 0xff; // Pullup
	
	// ADC
		ADCSRA = (1 << ADEN) | (1 << ADPS2) | (1 << ADPS1) | (0 << ADPS0);

	// Some global variables
	lastKey[0] = 0x0;
	lastKey[1] = 0x0;
	iii = 0;
	ADCidx = 0;
	SYSidx = 0;
}



void doKey(uchar keyPin, uchar ID) {
uchar idx;

	key[ID] = keyPin;

	if (key[ID] != lastKey[ID]) {
		for (idx = 0; idx < 8; idx++) { // Compare new key bitwise with the last state
			if ( (key[ID] & (1 << idx)) != (lastKey[ID] & (1 << idx))) {
				// Key change
				midiMsg[iii++] = (!(key[ID] & (1 << idx)) ? 0x08 : 0x09);
				midiMsg[iii++] = (!(key[ID] & (1 << idx)) ? 0x80 : 0x90);
				midiMsg[iii++] = idx + 41 + (ID ? 8 : 0);
				midiMsg[iii++] = (!(key[ID] & (1 << idx)) ? 0x00 : 0x7f);
				if ((key[ID] & (1 << idx))) lastKey[ID] |= (1 << idx);
						else	   lastKey[ID] &= ~(1 << idx);
				if (8 == iii) break; // MIDI Buffer is full
			}
		}
		
		//lastKey[ID] = key[ID];	
	}
}

void checkKeys(void) {
	iii = 0;
	
	doKey(~BTN2_PIN, 0); 

	if (8 != iii) { // If the Buffer isn't already full, read the second port.
		doKey((BTN1_PIN & 0b00001111) ^ (0b00001100), 1); // Only the first four bits are switches (and some of 'em are the wrong ones)
	}

	if (8 == iii) sendEmptyFrame = 1;
	else sendEmptyFrame = 0;
	if (iii > 0) usbSetInterrupt(midiMsg, iii);
}

#define ADC_READ_CYCELS 5
uint16_t readADC(uint8_t channel) {
	uint16_t result;
	uint8_t i;

	//ADC aktiv, Prescaler 16
	ADCSRA = (1 << ADEN) | (1 << ADPS2);

	//ADC0 kein mutiplexing
	ADMUX = channel;

	// AVcc als Versorgungspannung
	ADMUX |= (1 << REFS0);

	//ADC starten mit Dummyreadout
	ADCSRA |= (1 << ADSC);
	while(ADCSRA & (1 << ADSC));

	result = 0;

	for(i = 0; i < ADC_READ_CYCELS; i++) {
		ADCSRA |= (1 << ADSC);
		while(ADCSRA & (1 << ADSC));
		result += ADCW / ADC_READ_CYCELS;
		_delay_us(1);
	}

	//ADC aus
	ADCSRA &= ~(1 << ADEN);


	return result;
}

void doLED(uchar ID, uchar value) {
	if (ID == 1) {
		if (value == 64) LED_PORT |= (1 << LED_RIGHT);
		else LED_PORT &= ~(1 << LED_RIGHT);
	}
	else if (ID == 6) {
		if (value == 64) LED_PORT |= (1 << LED_LEFT);
		else LED_PORT &= ~(1 << LED_LEFT);
	}
	else if (ID == 4) {
		if (value == 64) LED_PORT |= (1 << LED_MIDDLE);
		else LED_PORT &= ~(1 << LED_MIDDLE);
	}
}

void checkAnalog() {

	uint16_t res;
	res = readADC(ADCidx);
	res = (res >> 3);
	if (res != oldADC[ADCidx]) {
		oldADC[ADCidx] = res;
		midiMsg[0] = 0x0b;
		midiMsg[1] = 0xb0;
		midiMsg[2] = ADCidx + 70;	// cc 70..77 
		midiMsg[3] = res;
		doLED(ADCidx, res);
		usbSetInterrupt(midiMsg, 4);
	}
	ADCidx++;
	if (ADCidx == 8) ADCidx = 0;
	
}

void checkSerial() {
	uint8_t uartval = uart_getc_nowait();
	if (uartval != 0xff) {
		uart_putc(uartval);
		if (uartval == 254) {
			serIdx=0;
		}
		else {
			serCode[serIdx] = uartval;
			serIdx++;
			if (serIdx >= 4) {
				midiMsg[0] = serCode[0];
				midiMsg[1] = serCode[1];
				midiMsg[2] = serCode[2];
				midiMsg[3] = serCode[3];
				sendEmptyFrame = 0;
				newSerial = 1;
			}
		}
	}
}

int main(void) {
	hardware_init();
	usbInit();
	uart_init(19600);

	wdt_enable(WDTO_1S);
	
	sei();

	while (1) {
		wdt_reset(); // Reset Watchdog
		usbPoll(); // Do some USB stuff
		
		checkSerial();

		if (usbInterruptIsReady()) {
			if (newSerial) { usbSetInterrupt(midiMsg, 4); newSerial=0; }
			else if (SYSidx == 0) checkKeys();
			else if (SYSidx == 1) checkAnalog();
			else SYSidx = -1;
			SYSidx++;
		}
	}
}


